class CpfGeneratorError(Exception):
    """Base exception for all cpf-gen related errors."""


class CpfGeneratorPrefixLengthError(CpfGeneratorError):
    """Raised when the prefix length is too long."""

    def __init__(self, prefix_length: int, max_length: int) -> None:
        self.prefix_length = prefix_length
        self.max_length = max_length

        super().__init__(
            f"The prefix length must be less than or equal to {max_length}. Got {prefix_length}."
        )


class CpfGeneratorPrefixNotValidError(CpfGeneratorError):
    """Raised when the prefix is not valid (e.g., repeated digits)."""

    def __init__(self, actual_prefix: str | list[str] | list[int], reason: str) -> None:
        self.actual_prefix = actual_prefix
        self.reason = reason

        super().__init__(f'The prefix "{actual_prefix}" is invalid. {reason}')



import random

from cpf_cd import CpfCheckDigits

from .cpf_generator_options import CpfGeneratorOptions
from .exceptions import CpfGeneratorPrefixNotValidError


class CpfGenerator:
    """Class to generate a valid CPF according to the given options."""

    __slots__ = "_options"

    def __init__(self, format: bool | None = None, prefix: str | None = None):
        self._options = CpfGeneratorOptions(format, prefix)

    def generate(self, format: bool | None = None, prefix: str | None = None) -> str:
        """Executes the CPF generation, overriding any given options with the ones set on the generator instance."""
        actual_options = self._options.merge(format, prefix)

        prefix_numbers = [int(digit) for digit in actual_options.prefix]
        cpf_sequence = self._generate_id(prefix_numbers)

        try:
            CpfGeneratorOptions(prefix=cpf_sequence)
        except CpfGeneratorPrefixNotValidError:
            return self.generate(format, prefix)

        cpf_check_digits = CpfCheckDigits(cpf_sequence)
        cpf_generated = cpf_check_digits.to_string()

        if actual_options.format:
            return self._format(cpf_generated)

        return cpf_generated

    def _generate_id(self, prefix_numbers: list[int]) -> list[int]:
        id_length = 9

        id_start = prefix_numbers[:id_length]
        id_start_length = len(id_start)

        id_end = [random.randint(0, 9) for _ in range(id_length - id_start_length)]

        return id_start + id_end

    def _format(self, cpf_string: str) -> str:
        return (
            f"{cpf_string[0:3]}.{cpf_string[3:6]}.{cpf_string[6:9]}-{cpf_string[9:11]}"
        )

    @property
    def options(self) -> CpfGeneratorOptions:
        """Direct access to the options manager for the CPF generator."""
        return self._options



import re
from dataclasses import dataclass, replace

from .exceptions import CpfGeneratorPrefixLengthError, CpfGeneratorPrefixNotValidError

CPF_LENGTH = 11
PREFIX_MAX_LENGTH = CPF_LENGTH - 2


@dataclass(slots=True, frozen=False)
class CpfGeneratorOptions:
    """Class to manage and store the options for the CPF generator."""

    format: bool | None = None
    prefix: str | None = None

    def __post_init__(self) -> None:
        if self.format is None:
            object.__setattr__(self, "format", False)
        if self.prefix is None:
            object.__setattr__(self, "prefix", "")

    def merge(
        self,
        format: bool | None = None,
        prefix: str | None = None,
    ) -> "CpfGeneratorOptions":
        """Creates a new instance of CpfGeneratorOptions with the given options merged with the current options."""
        kwargs = {}

        if format is not None:
            kwargs["format"] = format
        if prefix is not None:
            kwargs["prefix"] = prefix

        return replace(self, **kwargs)

    def __setattr__(self, name: str, value: object) -> None:
        if name == "prefix" and value is not None:
            prefix_value = re.sub(r"[^0-9]", "", str(value))
            prefix_length = len(prefix_value)

            if prefix_length > PREFIX_MAX_LENGTH:
                raise CpfGeneratorPrefixLengthError(prefix_length, PREFIX_MAX_LENGTH)

            if prefix_length == PREFIX_MAX_LENGTH:
                digits_set = set(prefix_value)

                if len(digits_set) == 1:
                    raise CpfGeneratorPrefixNotValidError(
                        value,
                        "Repeated digits are not considered valid.",
                    )

            value = prefix_value

        object.__setattr__(self, name, value)



from .cpf_gen import cpf_gen
from .cpf_generator import CpfGenerator
from .cpf_generator_options import CpfGeneratorOptions
from .exceptions import (
    CpfGeneratorError,
    CpfGeneratorPrefixLengthError,
    CpfGeneratorPrefixNotValidError,
)

__all__ = [
    "CpfGenerator",
    "CpfGeneratorError",
    "CpfGeneratorOptions",
    "CpfGeneratorPrefixLengthError",
    "CpfGeneratorPrefixNotValidError",
    "cpf_gen",
]

__version__ = "1.0.0"



from .cpf_generator import CpfGenerator


def cpf_gen(format: bool | None = None, prefix: str | None = None) -> str:
    """Generate a valid CPF according to the given options."""
    generator = CpfGenerator(format, prefix)

    return generator.generate()
